# coding: utf-8

import io
import os.path
from pkg_resources import parse_version
import base64
from typing import List, Tuple, Optional, Union
import cv2
from PIL import ImageDraw, ImageFile, ImageFont, Image as PILImage
import numpy as np
from enum import Enum
import nrrd

from supervisely.io.fs import ensure_base_path, get_file_ext, silent_remove
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.image_rotator import ImageRotator
from supervisely.imaging.font import get_font
from supervisely._utils import get_bytes_hash, is_development, abs_url, rand_str

ImageFile.LOAD_TRUNCATED_IMAGES = True

# @TODO: refactoring image->img
KEEP_ASPECT_RATIO = -1  # TODO: need move it to best place

# Do NOT use directly for image extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
SUPPORTED_IMG_EXTS = [".jpg", ".jpeg", ".mpo", ".bmp", ".png", ".webp", ".tiff", ".tif", ".nrrd"]
DEFAULT_IMG_EXT = ".png"


class CornerAnchorMode:
    """ """

    TOP_LEFT = "tl"
    TOP_RIGHT = "tr"
    BOTTOM_LEFT = "bl"
    BOTTOM_RIGHT = "br"


class RotateMode(Enum):
    """ """

    KEEP_BLACK = 0
    """"""
    CROP_BLACK = 1
    """"""
    SAVE_ORIGINAL_SIZE = 2
    """"""


class ImageExtensionError(Exception):
    """ """

    pass


class UnsupportedImageFormat(Exception):
    """ """

    pass


class ImageReadException(Exception):
    """ """

    pass


def is_valid_ext(ext: str) -> bool:
    """
    Checks file extension for list of supported images extensions('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp', '.tiff', '.tif', '.nrrd').

    :param ext: Image extention.
    :type ext: str
    :return: True if image extention in list of supported images extensions, False - in otherwise
    :rtype: :class:`bool`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        sly.image.is_valid_ext('.png') # True
        sly.image.is_valid_ext('.py') # False
    """
    return ext.lower() in SUPPORTED_IMG_EXTS


def has_valid_ext(path: str) -> bool:
    """
    Checks if a given file has a supported extension('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp', '.tiff', '.tif', '.nrrd').

    :param path: Path to file.
    :type path: str
    :return: True if file extention in list of supported images extensions, False - in otherwise
    :rtype: :class:`bool`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        sly.image.has_valid_ext('/home/admin/work/docs/new_image.jpeg') # True
        sly.image.has_valid_ext('/home/admin/work/docs/016_img.py') # False
    """
    return is_valid_ext(os.path.splitext(path)[1])


def validate_ext(path: str) -> None:
    """
    Generate exception error if file extention is not in list of supported images extensions('.jpg', '.jpeg', '.mpo', '.bmp', '.png', '.webp', '.tiff', '.tif', '.nrrd').

    :param path: Path to file.
    :type path: str
    :return: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        print(sly.image.validate_ext('/home/admin/work/docs/new_image.jpeg'))
        # Output: None

        try:
            print(sly.image.validate_ext('/home/admin/work/docs/016_img.py'))
        except ImageExtensionError as error:
            print(error)

        # Output: Unsupported image extension: '.py' for file '/home/admin/work/docs/016_img.py'. Only the following extensions are supported: .jpg, .jpeg, .mpo, .bmp, .png, .webp.
    """
    _, ext = os.path.splitext(path)
    if not is_valid_ext(ext):
        raise ImageExtensionError(
            "Unsupported image extension: {!r} for file {!r}. Only the following extensions are supported: {}.".format(
                ext, path, ", ".join(SUPPORTED_IMG_EXTS)
            )
        )


def validate_format(path: str) -> None:
    """
    Validate input file format, if file extension is not supported raise ImageExtensionError.

    :param path: Path to file.
    :type path: str
    :return: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        print(sly.image.validate_format('/home/admin/work/docs/new_image.jpeg'))
        # Output: None

        try:
            print(sly.image.validate_format('/home/admin/work/docs/016_img.py'))
        except ImageReadException as error:
            print(error)

        # Output: Error has occured trying to read image '/home/admin/work/docs/016_img.py'. Original exception message: "cannot identify image file '/home/admin/work/docs/016_img.py'"
    """
    ext = get_file_ext(path)
    if ext == ".nrrd":
        data, header = nrrd.read(path, index_order='C')
        return
    try:
        pil_img = PILImage.open(path)
        pil_img.load()  # Validate image data. Because 'open' is lazy method.
    except OSError as e:
        raise ImageReadException(
            "Error has occured trying to read image {!r}. Original exception message: {!r}".format(
                path, str(e)
            )
        )

    img_format = pil_img.format
    img_ext = f".{img_format}"
    if not is_valid_ext(f".{img_format}"):
        raise UnsupportedImageFormat(
            "Unsupported image format {!r} for file {!r}. Only the following formats are supported: {}".format(
                img_ext, path, ", ".join(SUPPORTED_IMG_EXTS)
            )
        )


def read(path: str, remove_alpha_channel: Optional[bool] = True) -> np.ndarray:
    """
    Loads an image from the specified file and returns it in RGB format.

    :param path: Path to file.
    :type path: str
    :param remove_alpha_channel: Define remove alpha channel when reading file or not.
    :type remove_alpha_channel: bool, optional
    :return: Numpy array
    :rtype: :class:`np.ndarray`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        im = sly.image.read('/home/admin/work/docs/image.jpeg')
    """
    ext = get_file_ext(path)
    if ext == ".nrrd":
        data, header = nrrd.read(path, index_order='C')
        return data

    validate_format(path)
    if remove_alpha_channel is True:
        img = cv2.imread(path, cv2.IMREAD_COLOR)
        if img is None:
            raise IOError("OpenCV can not open the file {!r}".format(path))
        return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    else:
        img = cv2.imread(path, cv2.IMREAD_UNCHANGED)
        if img is None:
            raise IOError("OpenCV can not open the file {!r}".format(path))
        cnt_channels = img.shape[2]
        if cnt_channels == 4:
            return cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
        elif cnt_channels == 3:
            return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        elif cnt_channels == 1:
            return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        else:
            raise ValueError(
                "image has {} channels. Please, contact support...".format(cnt_channels)
            )


def read_bytes(image_bytes: str, keep_alpha: Optional[bool] = False) -> np.ndarray:
    """
    Loads an byte image and returns it in RGB format.

    :param image_bytes: Path to file.
    :type image_bytes: str
    :param keep_alpha: Define consider alpha channel when reading bytes or not.
    :type keep_alpha: bool, optional
    :return: Numpy array
    :rtype: :class:`np.ndarray`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        im_bytes = '\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x01\x00\x00\x01\x00\x01\...\xd9'
        im = sly.image.read_bytes(im_bytes)
    """
    if image_bytes.startswith(b"NRRD"):
        file_like = io.BytesIO(image_bytes)
        header = nrrd.read_header(file_like)
        data = nrrd.read_data(header, file_like, index_order='C')
        return data

    image_np_arr = np.asarray(bytearray(image_bytes), dtype="uint8")
    if keep_alpha is True:
        img = cv2.imdecode(image_np_arr, cv2.IMREAD_UNCHANGED)
        if len(img.shape) == 2:
            img = np.expand_dims(img, 2)
        cnt_channels = img.shape[2]
        if cnt_channels == 4:
            img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
        elif cnt_channels == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        elif cnt_channels == 1:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        return img
    else:
        img = cv2.imdecode(
            image_np_arr, cv2.IMREAD_COLOR
        )  # cv2.imdecode returns BGR always
        return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)


def write(
    path: str, img: np.ndarray, remove_alpha_channel: Optional[bool] = True
) -> None:
    """
    Saves the image to the specified file. It create directory from path if the directory for this path does not exist.

    :param path: Path to file.
    :type path: str
    :param img: Image in numpy array(RGB format).
    :type img: np.ndarray
    :param remove_alpha_channel: Define remove alpha channel when writing file or not.
    :type remove_alpha_channel: bool, optional
    :return: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        path = '/home/admin/work/docs/new_image.jpeg'
        sly.image.write(path, image_np)
    """
    ensure_base_path(path)
    validate_ext(path)

    ext = get_file_ext(path)
    if ext == ".nrrd":
        return nrrd.write(path, img, index_order='C')

    res_img = img.copy()
    if len(img.shape) == 2:
        res_img = np.expand_dims(img, 2)
    cnt_channels = res_img.shape[2]
    if cnt_channels == 4:
        if remove_alpha_channel is True:
            res_img = cv2.cvtColor(res_img.astype(np.uint8), cv2.COLOR_RGBA2BGR)
        else:
            res_img = cv2.cvtColor(res_img.astype(np.uint8), cv2.COLOR_RGBA2BGRA)
    elif cnt_channels == 3:
        res_img = cv2.cvtColor(res_img.astype(np.uint8), cv2.COLOR_RGB2BGR)
    elif cnt_channels == 1:
        res_img = cv2.cvtColor(res_img.astype(np.uint8), cv2.COLOR_GRAY2BGR)
    return cv2.imwrite(path, res_img)


def draw_text_sequence(
    bitmap: np.ndarray,
    texts: List[str],
    anchor_point: Tuple[int, int],
    corner_snap: Optional[CornerAnchorMode] = CornerAnchorMode.TOP_LEFT,
    col_space: Optional[int] = 12,
    font: Optional[ImageFont.FreeTypeFont] = None,
    fill_background: Optional[bool] = True,
) -> None:
    """
    Draws text labels on bitmap from left to right with col_space spacing between labels.

    :param bitmap: Image to draw texts in numpy format.
    :type bitmap: np.ndarray
    :param texts: List of texts to draw on image.
    :type texts: List[str]
    :param anchor_point: Coordinates of the place on the image where the text will be displayed(row, column).
    :type anchor_point: Tuple[int, int]
    :param corner_snap: Corner of image to draw texts.
    :type corner_snap: CornerAnchorMode, optional
    :param col_space: Distance between texts.
    :type col_space: int, optional
    :param font: Type of text font.
    :type font: ImageFont.FreeTypeFont, optional
    :param fill_background: Define fill text background or not.
    :type fill_background: bool, optional
    :return: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        sly.image.draw_text_sequence(image, ['some_text', 'another_text'], (10, 10))

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/wIzrDuf.jpg

                   After
    """
    col_offset = 0
    for text in texts:
        position = anchor_point[0], anchor_point[1] + col_offset
        _, text_width = draw_text(
            bitmap, text, position, corner_snap, font, fill_background
        )
        col_offset += text_width + col_space


def draw_text(
    bitmap: np.ndarray,
    text: str,
    anchor_point: Tuple[int, int],
    corner_snap: Optional[CornerAnchorMode] = CornerAnchorMode.TOP_LEFT,
    font: Optional[ImageFont.FreeTypeFont] = None,
    fill_background: Optional[bool] = True,
    color: Optional[Union[Tuple[int, int, int, int], Tuple[int, int, int]]] = (0, 0, 0, 255),
) -> tuple:
    """
    Draws given text on bitmap image.

    :param bitmap: Image to draw texts in numpy format.
    :type bitmap: np.ndarray
    :param texts: Text to draw on image.
    :type texts: str
    :param anchor_point: Coordinates of the place on the image where the text will be displayed(row, column).
    :type anchor_point: Tuple[int, int]
    :param corner_snap: Corner of image to draw texts.
    :type corner_snap: CornerAnchorMode, optional
    :param font: Type of text font.
    :type font: ImageFont.FreeTypeFont, optional
    :param fill_background: Define fill text background or not.
    :type fill_background: bool, optional
    :param color: Text color as a tuple of three or four integers (red, green, blue, alpha)
                  ranging from 0 to 255.
                  If alpha is not provided, it defaults to 255 (fully opaque).
    :type color: Union[Tuple[int, int, int, int], Tuple[int, int, int]], optional
    :return: Height and width of text
    :rtype: :class:`Tuple[int, int]`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        sly.image.draw_text(image, 'your text', (100, 50), color=(0, 0, 0, 255))

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/6Ptz8Tf.jpg

                   After
    """
    if not (isinstance(color, (list, tuple)) and len(color) in [3, 4]):
        raise TypeError("Color must be list or tuple of three or four elements")

    if font is None:
        font = get_font()

    source_img = PILImage.fromarray(bitmap)
    source_img = source_img.convert("RGBA")

    canvas = PILImage.new("RGBA", source_img.size, (0, 0, 0, 0))
    drawer = ImageDraw.Draw(canvas, "RGBA")
    text_width, text_height = drawer.textsize(text, font=font)
    rect_top, rect_left = anchor_point

    if corner_snap == CornerAnchorMode.TOP_LEFT:
        pass  # Do nothing
    elif corner_snap == CornerAnchorMode.TOP_RIGHT:
        rect_left -= text_width
    elif corner_snap == CornerAnchorMode.BOTTOM_LEFT:
        rect_top -= text_height
    elif corner_snap == CornerAnchorMode.BOTTOM_RIGHT:
        rect_top -= text_height
        rect_left -= text_width

    if fill_background:
        rect_right = rect_left + text_width
        rect_bottom = rect_top + text_height
        drawer.rectangle(
            ((rect_left, rect_top), (rect_right + 1, rect_bottom)),
            fill=(255, 255, 255, 128),
        )
    drawer.text((rect_left + 1, rect_top), text, fill=color, font=font)

    source_img = PILImage.alpha_composite(source_img, canvas)
    source_img = source_img.convert("RGB")
    bitmap[:, :, :] = np.array(source_img, dtype=np.uint8)

    return (text_height, text_width)


# @TODO: not working with alpha channel
def write_bytes(img: np.ndarray, ext: str) -> bytes:
    """
    Compresses the image and stores it in the byte object.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :param ext: File extension that defines the output format.
    :type ext: str
    :return: Bytes object
    :rtype: :class:`bytes`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        bytes = sly.image.write_bytes(image_np, 'jpeg')
        print(type(bytes))
        # Output: <class 'bytes'>
    """
    ext = ("." + ext).replace("..", ".")
    if not is_valid_ext(ext):
        raise UnsupportedImageFormat(
            "Unsupported image format {!r}. Only the following formats are supported: {}".format(
                ext, ", ".join(SUPPORTED_IMG_EXTS)
            )
        )
    if ext == ".nrrd":
        nrrd_bytes = None
        _filename = f"./sly-nrrd-data-bytes-{rand_str(10)}{ext}"
        nrrd.write(_filename, img, index_order='C')
        with open(_filename, "rb") as nrrd_file:
            nrrd_bytes = nrrd_file.read()
        silent_remove(_filename)
        return nrrd_bytes

    img = cv2.cvtColor(img.astype(np.uint8), cv2.COLOR_RGB2BGR)
    encode_status, img_array = cv2.imencode(ext, img)
    if encode_status is True:
        return img_array.tobytes()
    raise RuntimeError("Can not encode input image")


# @TODO: not working with alpha channel
def get_hash(img: np.ndarray, ext: str) -> str:
    """
    Hash input image with sha256 algoritm and encode result by using Base64.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :param ext: File extension that defines the output format.
    :type ext: str
    :return: Hash string
    :rtype: :class:`str`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        hash = sly.image.get_hash(im, 'jpeg')
        print(hash)
        # Output: fTec3RD7Zxg0aYc0ooa5phPfBrzDe01urlFsgi5IzIQ=
    """
    return get_bytes_hash(write_bytes(img, ext))


def crop(img: np.ndarray, rect: Rectangle) -> np.ndarray:
    """
    Crop part of the image with rectangle size. If rectangle for crop is out of image area it generates ValueError.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :param rect: :class:`Rectangle<supervisely.geometry.rectangle.Rectangle>` object for crop.
    :type rect: Rectangle
    :return: Cropped image in numpy format
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # If size of rectangle is more then image shape raise ValueError:
        try:
            crop_image = sly.image.crop(image_np, sly.Rectangle(0, 0, 5000, 6000))
        except ValueError as error:
            print(error)
        # Output: Rectangle for crop out of image area!

        crop_im = sly.image.crop(image_np, sly.Rectangle(0, 0, 500, 600))

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/4tNm2GS.jpg

                   After
    """
    img_rect = Rectangle.from_array(img)
    if not img_rect.contains(rect):
        raise ValueError("Rectangle for crop out of image area!")
    return rect.get_cropped_numpy_slice(img)


def crop_with_padding(img: np.ndarray, rect: Rectangle) -> np.ndarray:
    """
    Crop part of the image with rectangle size. If rectangle for crop is out of image area it generates additional padding.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :param rect: :class:`Rectangle<supervisely.geometry.rectangle.Rectangle>` object for crop.
    :type rect: Rectangle
    :return: Cropped image in numpy format
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        crop_with_padding_image = sly.image.crop_with_padding(image_np, sly.Rectangle(0, 0, 1000, 1200))

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/Nv1UinH.jpg

                   After
    """
    img_rect = Rectangle.from_array(img)
    if not img_rect.contains(rect):
        row, col = img.shape[:2]
        new_img = cv2.copyMakeBorder(
            img,
            top=rect.height,
            bottom=rect.height,
            left=rect.width,
            right=rect.width,
            borderType=cv2.BORDER_CONSTANT,
        )
        new_rect = rect.translate(drow=rect.height, dcol=rect.width)
        return new_rect.get_cropped_numpy_slice(new_img)

    else:
        return rect.get_cropped_numpy_slice(img)


def restore_proportional_size(
    in_size: Tuple[int, int],
    out_size: Optional[Tuple[int, int]] = None,
    frow: Optional[float] = None,
    fcol: Optional[float] = None,
    f: Optional[float] = None,
) -> Tuple[int, int]:
    """
    Calculate new size of the image.

    :param in_size: Size of input image (height, width).
    :type in_size: Tuple[int, int]
    :param out_size: New image size (height, width).
    :type out_size: Tuple[int, int], optional
    :param frow: Length of output image.
    :type frow: float, optional
    :param fcol: Height of output image.
    :type fcol: float, optional
    :param f: Positive non zero scale factor.
    :type f: float, optional
    :return: Height and width of image
    :rtype: :class:`Tuple[int, int]`
    """
    if out_size is not None and (frow is not None or fcol is not None) and f is None:
        raise ValueError(
            "Must be specified output size or scale factors not both of them."
        )

    if out_size is not None:
        if out_size[0] == KEEP_ASPECT_RATIO and out_size[1] == KEEP_ASPECT_RATIO:
            raise ValueError("Must be specified at least 1 dimension of size!")

        if (out_size[0] <= 0 and out_size[0] != KEEP_ASPECT_RATIO) or (
            out_size[1] <= 0 and out_size[1] != KEEP_ASPECT_RATIO
        ):
            raise ValueError("Size dimensions must be greater than 0.")

        result_row = (
            out_size[0]
            if out_size[0] > 0
            else max(1, round(out_size[1] / in_size[1] * in_size[0]))
        )
        result_col = (
            out_size[1]
            if out_size[1] > 0
            else max(1, round(out_size[0] / in_size[0] * in_size[1]))
        )
    else:
        if f is not None:
            if f < 0:
                raise ValueError('"f" argument must be positive!')
            frow = fcol = f

        if (frow < 0 or fcol < 0) or (frow is None or fcol is None):
            raise ValueError('Specify "f" argument for single scale!')

        result_col = round(fcol * in_size[1])
        result_row = round(frow * in_size[0])
    return result_row, result_col


# @TODO: reimplement, to be more convenient
def resize(
    img: np.ndarray,
    out_size: Optional[Tuple[int, int]] = None,
    frow: Optional[float] = None,
    fcol: Optional[float] = None,
) -> np.ndarray:
    """
    Resize the image to the specified size.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :param out_size: New image size (height, width).
    :type out_size: Tuple[int, int], optional
    :param frow: Length of output image.
    :type frow: float, optional
    :param fcol: Height of output image.
    :type fcol: float, optional
    :return: Resize image in numpy format
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        resize_image = sly.image.resize(image_np, (300, 500))

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/Xya4yz0.jpg

                   After
    """
    result_height, result_width = restore_proportional_size(
        img.shape[:2], out_size, frow, fcol
    )
    return cv2.resize(img, (result_width, result_height), interpolation=cv2.INTER_CUBIC)


def resize_inter_nearest(
    img: np.ndarray,
    out_size: Optional[Tuple[int, int]] = None,
    frow: Optional[float] = None,
    fcol: Optional[float] = None,
) -> np.ndarray:
    """
    Resize image to match a certain size. Performs interpolation to up-size or down-size images.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :param out_size: New image size (height, width).
    :type out_size: Tuple[int, int], optional
    :param frow: Length of output image.
    :type frow: float, optional
    :param fcol: Height of output image.
    :type fcol: float, optional
    :return: Resize image in numpy format
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        resize_image_nearest = sly.image.resize_inter_nearest(image_np, (300, 700))

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/0O6yMDH.jpg

                   After
    """
    import skimage.transform

    target_shape = restore_proportional_size(img.shape[:2], out_size, frow, fcol)
    resize_kv_args = dict(order=0, preserve_range=True, mode="constant")
    if parse_version(skimage.__version__) >= parse_version("0.14.0"):
        resize_kv_args["anti_aliasing"] = False
    return skimage.transform.resize(img, target_shape, **resize_kv_args).astype(
        img.dtype
    )


def scale(img: np.ndarray, factor: float) -> np.ndarray:
    """
    Scales current image with the given factor.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :param factor: Scale size.
    :type factor: float
    :return: Resize image in numpy format
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        scale_image = sly.image.scale(image_np, 0.3)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/NyP8tts.jpg

                   After
    """
    return resize(img, (round(img.shape[0] * factor), round(img.shape[1] * factor)))


def fliplr(img: np.ndarray) -> np.ndarray:
    """
    Flips the current image horizontally.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :return: Flip image in numpy format
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        fliplr_image = sly.image.fliplr(image_np)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/1mqnuZU.jpg

                   After
    """
    return np.flip(img, 1)


def flipud(img: np.ndarray) -> np.ndarray:
    """
    Flips the current image vertically.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :return: Flip image in numpy format
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        flipud_image = sly.image.flipud(image_np)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/LDwRDvm.jpg

                   After
    """
    return np.flip(img, 0)


def rotate(
    img: np.ndarray,
    degrees_angle: int,
    mode: Optional[RotateMode] = RotateMode.KEEP_BLACK,
) -> np.ndarray:
    """
    Rotates current image.

    :param img: Image in numpy format(RGB).
    :type img: np.ndarray
    :param degrees_angle: Angle in degrees for rotating.
    :type degrees_angle: int
    :param mode: One of RotateMode enum values.
    :type mode: RotateMode, optional
    :return: Rotate image in numpy format
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # keep_black mode
        rotate_im_keep_black = sly.image.rotate(image_np, 45)
        # crop_black mode
        rotate_im_crop_black = sly.image.rotate(image_np, 45, sly.image.RotateMode.CROP_BLACK)
        # origin_size mode
        rotate_im_origin_size = sly.image.rotate(image_np, 45, sly.image.RotateMode.SAVE_ORIGINAL_SIZE) * 255

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/VjiwV4O.jpg

                   After keep_black mode

          - .. figure:: https://i.imgur.com/Rs34eMa.jpg

                   After crop_black mode

          - .. figure:: https://i.imgur.com/ttDWfBE.jpg

                   After origin_size mode
    """
    import skimage.transform

    rotator = ImageRotator(imsize=img.shape[:2], angle_degrees_ccw=degrees_angle)
    if mode == RotateMode.KEEP_BLACK:
        return rotator.rotate_img(img, use_inter_nearest=False)  # @TODO: order = ???
    elif mode == RotateMode.CROP_BLACK:
        img_rotated = rotator.rotate_img(img, use_inter_nearest=False)
        return rotator.inner_crop.get_cropped_numpy_slice(img_rotated)
    elif mode == RotateMode.SAVE_ORIGINAL_SIZE:
        # TODO Implement this in rotator instead.
        return skimage.transform.rotate(img, degrees_angle, resize=False)
    else:
        raise NotImplementedError('Rotate mode "{0}" not supported!'.format(str(mode)))


# Color augmentations
def _check_contrast_brightness_inputs(min_value, max_value):
    """
    The function _check_contrast_brightness_inputs checks the input brightness or contrast for correctness.
    """
    if min_value < 0:
        raise ValueError("Minimum value must be greater than or equal to 0.")
    if min_value > max_value:
        raise ValueError(
            "Maximum value must be greater than or equal to minimum value."
        )


def random_contrast(
    image: np.ndarray, min_factor: float, max_factor: float
) -> np.ndarray:
    """
    Randomly changes contrast of the input image.

    :param image: Image in numpy format(RGB).
    :type image: np.ndarray
    :param min_factor: Lower bound of contrast range.
    :type min_factor: float
    :param max_factor: Upper bound of contrast range.
    :type max_factor: float
    :return: Image in numpy format with new contrast
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        rand_contrast_im = sly.image.random_contrast(image_np, 1.1, 1.8)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/4zSNuJU.jpg

                   After
    """
    _check_contrast_brightness_inputs(min_factor, max_factor)
    contrast_value = np.random.uniform(min_factor, max_factor)
    image_gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    image_mean = round(image_gray.mean())
    image = image.astype(np.float32)
    image = contrast_value * (image - image_mean) + image_mean
    return np.clip(image, 0, 255).astype(np.uint8)


def random_brightness(
    image: np.ndarray, min_factor: float, max_factor: float
) -> np.ndarray:
    """
    Randomly changes brightness of the input image.

    :param image: Image in numpy format(RGB).
    :type image: np.ndarray
    :param min_factor: Lower bound of brightness range.
    :type min_factor: float
    :param max_factor: Upper bound of brightness range.
    :type max_factor: float
    :return: Image in numpy format with new brightness
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        rand_brightness_im = sly.image.random_brightness(image_np, 1.5, 8.5)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/bOYwwYH.jpg

                   After
    """
    _check_contrast_brightness_inputs(min_factor, max_factor)
    brightness_value = np.random.uniform(min_factor, max_factor)
    image = image.astype(np.float32)
    image = image * brightness_value
    return np.clip(image, 0, 255).astype(np.uint8)


def random_noise(image: np.ndarray, mean: float, std: float) -> np.ndarray:
    """
    Adds random Gaussian noise to the input image.

    :param image: Image in numpy format(RGB).
    :type image: np.ndarray
    :param mean: The mean value of noise distribution.
    :type mean: float
    :param std: The standard deviation of noise distribution.
    :type std: float
    :return: Image in numpy format with random noise
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        random_noise_im = sly.image.random_noise(image_np, 25, 19)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/EzyEHeM.jpg

                   After
    """
    image = image.astype(np.float32)
    image += np.random.normal(mean, std, image.shape)

    return np.clip(image, 0, 255).astype(np.uint8)


def random_color_scale(
    image: np.ndarray, min_factor: float, max_factor: float
) -> np.ndarray:
    """
    Changes image colors by randomly scaling each of RGB components. The scaling factors are sampled uniformly from the given range.

    :param image: Image in numpy format(RGB).
    :type image: np.ndarray
    :param min_factor: Minimum scale factor.
    :type min_factor: float
    :param max_factor: Maximum scale factor.
    :type max_factor: float
    :return: Image in numpy format with random color scale
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        random_color_scale_im = sly.image.random_color_scale(image_np, 0.5, 0.9)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/GGUZqlA.jpg

                   After
    """
    image_float = image.astype(np.float64)
    scales = np.random.uniform(
        low=min_factor, high=max_factor, size=(1, 1, image.shape[2])
    )
    res_image = image_float * scales
    return np.clip(res_image, 0, 255).astype(np.uint8)


# Blurs
def blur(image: np.ndarray, kernel_size: int) -> np.ndarray:
    """
    Blurs an image using the normalized box filter.

    :param image: Image in numpy format(RGB).
    :type image: np.ndarray
    :param kernel_size: Blurring kernel size.
    :type kernel_size: int
    :return: Image in numpy format with blur
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        blur_im = sly.image.blur(image_np, 7)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/wFnBaC6.jpg

                   After
    """
    return cv2.blur(image, (kernel_size, kernel_size))


def median_blur(image: np.ndarray, kernel_size: int) -> np.ndarray:
    """
    Blurs an image using the median filter.

    :param image: Image in numpy format(RGB).
    :type image: np.ndarray
    :param kernel_size: Blurring kernel size(must be odd and greater than 1, for example: 3, 5, 7).
    :type kernel_size: int
    :return: Image in numpy format with median blur
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        median_blur_im = sly.image.median_blur(image_np, 5)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/FQ977ON.jpg

                   After
    """
    return cv2.medianBlur(image, kernel_size)


def gaussian_blur(image: np.ndarray, sigma_min: float, sigma_max: float) -> np.ndarray:
    """
    Blurs an image using a Gaussian filter.

    :param image: Image in numpy format(RGB).
    :type image: np.ndarray
    :param sigma_min: Lower bound of Gaussian kernel standard deviation range.
    :type sigma_min: float
    :param sigma_min: Upper bound of Gaussian kernel standard deviation range.
    :type sigma_max: float
    :return: Image in numpy format with gaussian blur
    :rtype: :class:`np.ndarray`

    :Usage example:

     .. code-block:: python

        import supervisely as sly

        gaussian_blur_im = sly.image.gaussian_blur(image_np, 3.3, 7.5)

    .. list-table::

        * - .. figure:: https://i.imgur.com/BHUALdv.jpg

                   Before

          - .. figure:: https://i.imgur.com/brs6Au0.jpg

                   After
    """
    sigma_value = np.random.uniform(sigma_min, sigma_max)
    return cv2.GaussianBlur(image, (0, 0), sigma_value)


def drop_image_alpha_channel(img: np.ndarray) -> np.ndarray:
    """
    Converts 4-channel image to 3-channel.

    :param img: Image in numpy format(RGBA).
    :type img: np.ndarray
    :return: Image in numpy format(RGB)
    :rtype: :class:`np.ndarray`
    """
    if img.shape[2] != 4:
        raise ValueError(
            "Only 4-channel RGBA images are supported for alpha channel removal. "
            + "Instead got {} channels.".format(img.shape[2])
        )
    return cv2.cvtColor(img, cv2.COLOR_RGBA2RGB)


# @TODO: refactor from two separate methods to a single one
# bgra or bgr
def np_image_to_data_url(img: np.ndarray) -> str:
    """
    Convert image to url string.

    :param img: Image in numpy format(RGBA or RGB).
    :type img: np.ndarray
    :return: String with image url
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        data_url = sly.image.np_image_to_data_url(im)
        print(data_url)
        # Output: 'data:image/png;base64,iVBORw0K...'
    """
    encode_status, bgra_result_png = cv2.imencode(".png", img)
    img_png = bgra_result_png.tobytes()
    img_base64 = base64.b64encode(img_png)
    data_url = "data:image/png;base64,{}".format(str(img_base64, "utf-8"))
    return data_url


def data_url_to_numpy(data_url: str) -> np.ndarray:
    """
    Convert url string to numpy image(RGB).

    :param img: String with image url.
    :type img: str
    :return: Image in numpy format(RGBA)
    :rtype: :class:`np.ndarray`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        image_np = sly.image.data_url_to_numpy(data_url)
    """
    img_base64 = data_url[len("data:image/png;base64,") :]
    img_base64 = base64.b64decode(img_base64)
    image = read_bytes(img_base64)
    return image


# only rgb
def np_image_to_data_url_backup_rgb(img: np.ndarray) -> str:
    """
    Convert image to url string.

    :param img: Image in numpy format(only RGB).
    :type img: np.ndarray
    :return: String with image url
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        data_url = sly.image.np_image_to_data_url_backup_rgb(im)
        print(data_url)
        # Output: 'data:image/png;base64,iVBORw0K...'
    """
    img_base64 = base64.b64encode(write_bytes(img, "png"))
    data_url = "data:image/png;base64,{}".format(str(img_base64, "utf-8"))
    return data_url


def get_labeling_tool_url(team_id, workspace_id, project_id, dataset_id, image_id):
    res = f"/app/images/{team_id}/{workspace_id}/{project_id}/{dataset_id}#image-{image_id}"
    if is_development():
        res = abs_url(res)
    return res


def get_labeling_tool_link(url, name="open in labeling tool"):
    return f'<a href="{url}" rel="noopener noreferrer" target="_blank">{name}<i class="zmdi zmdi-open-in-new" style="margin-left: 5px"></i></a>'



# coding: utf-8
"""download/upload/edit :class:`Annotation<supervisely.annotation.annotation.Annotation>`"""

# docs
from __future__ import annotations
import json
from typing import List, Optional, NamedTuple, Dict, Callable, Union
from supervisely.task.progress import Progress
from supervisely.annotation.label import Label

from supervisely.annotation.annotation import Annotation
from supervisely.api.module_api import ApiField, ModuleApi
from supervisely._utils import batched
from tqdm import tqdm


class AnnotationInfo(NamedTuple):
    """
    AnnotationInfo
    """

    image_id: int
    image_name: str
    annotation: dict
    created_at: str
    updated_at: str


class AnnotationApi(ModuleApi):
    """
    Annotation for a single image. :class:`AnnotationApi<AnnotationApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        dataset_id = 254737
        ann_infos = api.annotation.get_list(dataset_id)
    """

    @staticmethod
    def info_sequence():
        """
        NamedTuple AnnotationInfo information about Annotation.

        :Example:

         .. code-block:: python

            AnnotationInfo(image_id=121236919,
                           image_name='IMG_1836',
                           annotation={'description': '', 'tags': [], 'size': {'height': 800, 'width': 1067}, 'objects': []},
                           created_at='2019-12-19T12:06:59.435Z',
                           updated_at='2021-02-06T11:07:26.080Z')
        """
        return [
            ApiField.IMAGE_ID,
            ApiField.IMAGE_NAME,
            ApiField.ANNOTATION,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **AnnotationInfo**.
        """
        return "AnnotationInfo"

    def get_list(
        self,
        dataset_id: int,
        filters: Optional[List[Dict[str, str]]] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        force_metadata_for_links: Optional[bool] = True,
    ) -> List[AnnotationInfo]:
        """
        Get list of information about all annotations for a given dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param filters: List of parameters to sort output Annotations.
        :type filters: List[dict], optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about Annotations. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[AnnotationInfo]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 254737
            ann_infos = api.annotation.get_list(dataset_id)
            print(json.dumps(ann_infos[0], indent=4))
            # Output: [
            #     121236918,
            #     "IMG_0748.jpeg",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]

            ann_infos_filter = api.annotation.get_list(dataset_id, filters={ 'field': 'name', 'operator': '=', 'value': 'IMG_1836' })
            print(json.dumps(ann_infos_filter, indent=4))
            # Output: [
            #     121236919,
            #     "IMG_1836",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]
        """
        return self.get_list_all_pages(
            "annotations.list",
            {
                ApiField.DATASET_ID: dataset_id,
                ApiField.FILTER: filters or [],
                ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
            },
            progress_cb,
        )

    def get_list_generator(
        self,
        dataset_id: int,
        filters: Optional[List[Dict[str, str]]] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        batch_size: Optional[int] = 50,
        force_metadata_for_links: Optional[bool] = True,
    ) -> List[AnnotationInfo]:
        """
        Get list of information about all annotations for a given dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param filters: List of parameters to sort output Annotations.
        :type filters: List[dict], optional
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: Information about Annotations. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[AnnotationInfo]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 254737
            ann_infos = api.annotation.get_list(dataset_id)
            print(json.dumps(ann_infos[0], indent=4))
            # Output: [
            #     121236918,
            #     "IMG_0748.jpeg",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]

            ann_infos_filter = api.annotation.get_list(dataset_id, filters={ 'field': 'name', 'operator': '=', 'value': 'IMG_1836' })
            print(json.dumps(ann_infos_filter, indent=4))
            # Output: [
            #     121236919,
            #     "IMG_1836",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]
        """
        data = {
            ApiField.DATASET_ID: dataset_id,
            ApiField.FILTER: filters or [],
            ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
            ApiField.PAGINATION_MODE: ApiField.TOKEN,
        }
        if batch_size is not None:
            data[ApiField.PER_PAGE] = batch_size
        else:
            # use default value on instance (learn in API documentation)
            # 20k for instance
            # 50 by default in SDK
            pass

        return self.get_list_all_pages_generator("annotations.list", data, progress_cb)

    def download(
        self,
        image_id: int,
        with_custom_data: Optional[bool] = False,
        force_metadata_for_links: Optional[bool] = True,
    ) -> AnnotationInfo:
        """
        Download AnnotationInfo by image ID from API.

        :param image_id: Image ID in Supervisely.
        :type image_id: int
        :param with_custom_data:
        :type with_custom_data: bool, optional
        :return: Information about Annotation. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`AnnotationInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            ann_info = api.annotation.download(image_id)
            print(json.dumps(ann_info, indent=4))
            # Output: [
            #     121236918,
            #     "IMG_0748.jpeg",
            #     {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     },
            #     "2019-12-19T12:06:59.435Z",
            #     "2021-02-06T11:07:26.080Z"
            # ]
        """
        response = self._api.post(
            "annotations.info",
            {
                ApiField.IMAGE_ID: image_id,
                ApiField.WITH_CUSTOM_DATA: with_custom_data,
                ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
            },
        )
        return self._convert_json_info(response.json())

    def download_json(
        self,
        image_id: int,
        with_custom_data: Optional[bool] = False,
        force_metadata_for_links: Optional[bool] = True,
    ) -> Dict[str, Union[str, int, list, dict]]:
        """
        Download Annotation in json format by image ID from API.

        :param image_id: Image ID in Supervisely.
        :type image_id: int
        :param with_custom_data:
        :type with_custom_data: bool, optional
        :return: Annotation in json format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            ann_json = api.annotation.download_json(image_id)
            print(ann_json)
            # Output: {
            #         "description": "",
            #         "tags": [],
            #         "size": {
            #             "height": 800,
            #             "width": 1067
            #         },
            #         "objects": []
            #     }
        """
        return self.download(
            image_id=image_id,
            with_custom_data=with_custom_data,
            force_metadata_for_links=force_metadata_for_links,
        ).annotation

    def download_batch(
        self,
        dataset_id: int,
        image_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        with_custom_data: Optional[bool] = False,
        force_metadata_for_links: Optional[bool] = True,
    ) -> List[AnnotationInfo]:
        """
        Get list of AnnotationInfos for given dataset ID from API.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: List of integers.
        :type image_ids: List[int]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm
        :return: Information about Annotations. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[AnnotationInfo]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 254737
            image_ids = [121236918, 121236919]
            p = tqdm(desc="Annotations downloaded: ", total=len(image_ids))

            ann_infos = api.annotation.download_batch(dataset_id, image_ids, progress_cb=p)
            # Output:
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations downloaded: ", "current": 0, "total": 2, "timestamp": "2021-03-16T15:20:06.168Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations downloaded: ", "current": 2, "total": 2, "timestamp": "2021-03-16T15:20:06.510Z", "level": "info"}
        """
        id_to_ann = {}
        for batch in batched(image_ids):
            post_data = {
                ApiField.DATASET_ID: dataset_id,
                ApiField.IMAGE_IDS: batch,
                ApiField.WITH_CUSTOM_DATA: with_custom_data,
                ApiField.FORCE_METADATA_FOR_LINKS: force_metadata_for_links,
            }
            results = self._api.post("annotations.bulk.info", data=post_data).json()
            for ann_dict in results:
                ann_info = self._convert_json_info(ann_dict)
                id_to_ann[ann_info.image_id] = ann_info
            if progress_cb is not None:
                progress_cb(len(batch))
        ordered_results = [id_to_ann[image_id] for image_id in image_ids]
        return ordered_results

    def download_json_batch(
        self,
        dataset_id: int,
        image_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        force_metadata_for_links: Optional[bool] = True,
    ) -> List[Dict]:
        """
        Get list of AnnotationInfos for given dataset ID from API.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param image_ids: List of integers.
        :type image_ids: List[int]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm
        :return: Information about Annotations. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[Dict]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 254737
            image_ids = [121236918, 121236919]
            p = tqdm(desc="Annotations downloaded: ", total=len(image_ids))

            anns_jsons = api.annotation.download_json_batch(dataset_id, image_ids, progress_cb=p)
            # Output:
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations downloaded: ", "current": 0, "total": 2, "timestamp": "2021-03-16T15:20:06.168Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations downloaded: ", "current": 2, "total": 2, "timestamp": "2021-03-16T15:20:06.510Z", "level": "info"}
        """
        results = self.download_batch(
            dataset_id=dataset_id,
            image_ids=image_ids,
            progress_cb=progress_cb,
            force_metadata_for_links=force_metadata_for_links,
        )
        return [ann_info.annotation for ann_info in results]

    def upload_path(
        self,
        img_id: int,
        ann_path: str,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an annotation from a given path to a given image ID in the API.

        :param img_id: Image ID in Supervisely.
        :type img_id: int
        :param ann_path: Path to annotation on host.
        :type ann_path: str
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            ann_path = '/home/admin/work/supervisely/example/ann.json'
            upl_path = api.annotation.upload_path(image_id, ann_path)
        """
        self.upload_paths([img_id], [ann_path], skip_bounds_validation=skip_bounds_validation)

    def upload_paths(
        self,
        img_ids: List[int],
        ann_paths: List[str],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an annotations from a given paths to a given images IDs in the API. Images IDs must be from one dataset.

        :param img_ids: Images IDs in Supervisely.
        :type img_ids: List[int]
        :param ann_paths: Paths to annotations on local machine.
        :type ann_paths: List[str]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_ids = [121236918, 121236919]
            ann_pathes = ['/home/admin/work/supervisely/example/ann1.json', '/home/admin/work/supervisely/example/ann2.json']
            upl_paths = api.annotation.upload_paths(img_ids, ann_pathes)
        """

        def read_json(ann_path):
            with open(ann_path) as json_file:
                return json.load(json_file)

        self._upload_batch(
            read_json,
            img_ids,
            ann_paths,
            progress_cb,
            skip_bounds_validation=skip_bounds_validation,
        )

    def upload_json(
        self,
        img_id: int,
        ann_json: Dict,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an annotation from dict to a given image ID in the API.

        :param img_id: Image ID in Supervisely.
        :type img_id: int
        :param ann_json: Annotation in JSON format.
        :type ann_json: dict
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            upl_json = api.annotation.upload_json(image_id, ann_json)
        """
        self.upload_jsons([img_id], [ann_json], skip_bounds_validation=skip_bounds_validation)

    def upload_jsons(
        self,
        img_ids: List[int],
        ann_jsons: List[Dict],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an annotations from dicts to a given images IDs in the API. Images IDs must be from one dataset.

        :param img_ids: Image ID in Supervisely.
        :type img_ids: List[int]
        :param ann_jsons: Annotation in JSON format.
        :type ann_jsons: List[dict]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_ids = [121236918, 121236919]
            upl_jsons = api.annotation.upload_jsons(img_ids, ann_jsons)
        """
        self._upload_batch(
            lambda x: x,
            img_ids,
            ann_jsons,
            progress_cb,
            skip_bounds_validation=skip_bounds_validation,
        )

    def upload_ann(
        self,
        img_id: int,
        ann: Annotation,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an :class:`Annotation<supervisely.annotation.annotation.Annotation>` to a given image ID in the API.

        :param img_id: Image ID in Supervisely.
        :type img_id: int
        :param ann: Annotation object.
        :type ann: Annotation
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            image_id = 121236918
            upl_ann = api.annotation.upload_ann(image_id, ann)
        """
        self.upload_anns([img_id], [ann], skip_bounds_validation=skip_bounds_validation)

    def upload_anns(
        self,
        img_ids: List[int],
        anns: List[Annotation],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Loads an :class:`Annotations<supervisely.annotation.annotation.Annotation>` to a given images IDs in the API. Images IDs must be from one dataset.

        :param img_ids: Image ID in Supervisely.
        :type img_ids: List[int]
        :param anns: List of Annotation objects.
        :type anns: List[Annotation]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_ids = [121236918, 121236919]
            upl_anns = api.annotation.upload_anns(img_ids, [ann1, ann2])
        """
        # img_ids from the same dataset
        self._upload_batch(
            Annotation.to_json,
            img_ids,
            anns,
            progress_cb,
            skip_bounds_validation=skip_bounds_validation,
        )

    def _upload_batch(
        self,
        func_ann_to_json,
        img_ids,
        anns,
        progress_cb=None,
        skip_bounds_validation: Optional[bool] = False,
    ):
        """
        _upload_batch
        """
        # img_ids from the same dataset
        if len(img_ids) == 0:
            return
        if len(img_ids) != len(anns):
            raise RuntimeError(
                'Can not match "img_ids" and "anns" lists, len(img_ids) != len(anns)'
            )

        dataset_id = self._api.image.get_info_by_id(
            img_ids[0], force_metadata_for_links=False
        ).dataset_id
        for batch in batched(list(zip(img_ids, anns))):
            data = [
                {ApiField.IMAGE_ID: img_id, ApiField.ANNOTATION: func_ann_to_json(ann)}
                for img_id, ann in batch
            ]
            self._api.post(
                "annotations.bulk.add",
                data={
                    ApiField.DATASET_ID: dataset_id,
                    ApiField.ANNOTATIONS: data,
                    ApiField.SKIP_BOUNDS_VALIDATION: skip_bounds_validation,
                },
            )
            if progress_cb is not None:
                progress_cb(len(batch))

    def get_info_by_id(self, id):
        """
        get_info_by_id
        """
        raise NotImplementedError("Method is not supported")

    def get_info_by_name(self, parent_id, name):
        """
        get_info_by_name
        """
        raise NotImplementedError("Method is not supported")

    def exists(self, parent_id, name):
        """
        exists
        """
        raise NotImplementedError("Method is not supported")

    def get_free_name(self, parent_id, name):
        """
        get_free_name
        """
        raise NotImplementedError("Method is not supported")

    def _add_sort_param(self, data):
        """
        _add_sort_param
        """
        return data

    def copy_batch(
        self,
        src_image_ids: List[int],
        dst_image_ids: List[int],
        progress_cb: Optional[Union[tqdm, Callable]] = None,
        force_metadata_for_links: Optional[bool] = True,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Copy annotations from one images IDs to another in API.

        :param src_image_ids: Images IDs in Supervisely.
        :type src_image_ids: List[int]
        :param dst_image_ids: Unique IDs of images in API.
        :type dst_image_ids: List[int]
        :param progress_cb: Function for tracking download progress.
        :type progress_cb: tqdm or callable, optional
        :raises: :class:`RuntimeError`, if len(src_image_ids) != len(dst_image_ids)
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from tqdm import tqdm

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_ids = [121236918, 121236919]
            dst_ids = [547837053, 547837054]
            p = tqdm(desc="Annotations copy: ", total=len(src_ids))

            copy_anns = api.annotation.copy_batch(src_ids, dst_ids, progress_cb=p)
            # Output:
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations copy: ", "current": 0, "total": 2, "timestamp": "2021-03-16T15:24:31.286Z", "level": "info"}
            # {"message": "progress", "event_type": "EventType.PROGRESS", "subtask": "Annotations copy: ", "current": 2, "total": 2, "timestamp": "2021-03-16T15:24:31.288Z", "level": "info"}
        """
        if len(src_image_ids) != len(dst_image_ids):
            raise RuntimeError(
                'Can not match "src_image_ids" and "dst_image_ids" lists, '
                "len(src_image_ids) != len(dst_image_ids)"
            )
        if len(src_image_ids) == 0:
            return

        src_dataset_id = self._api.image.get_info_by_id(src_image_ids[0]).dataset_id
        for cur_batch in batched(list(zip(src_image_ids, dst_image_ids))):
            src_ids_batch, dst_ids_batch = zip(*cur_batch)
            ann_infos = self.download_batch(
                src_dataset_id, src_ids_batch, force_metadata_for_links=force_metadata_for_links
            )
            ann_jsons = [ann_info.annotation for ann_info in ann_infos]
            self.upload_jsons(
                dst_ids_batch, ann_jsons, skip_bounds_validation=skip_bounds_validation
            )
            if progress_cb is not None:
                progress_cb(len(src_ids_batch))

    def copy(
        self,
        src_image_id: int,
        dst_image_id: int,
        force_metadata_for_links: Optional[bool] = True,
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Copy annotation from one image ID to another image ID in API.

        :param src_image_id: Image ID in Supervisely.
        :type src_image_id: int
        :param dst_image_id: Image ID in Supervisely.
        :type dst_image_id: int
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_id = 121236918
            dst_id = 547837053
            api.annotation.copy(src_id, dst_id)
        """
        self.copy_batch(
            [src_image_id],
            [dst_image_id],
            force_metadata_for_links=force_metadata_for_links,
            skip_bounds_validation=skip_bounds_validation,
        )

    def copy_batch_by_ids(
        self,
        src_image_ids: List[int],
        dst_image_ids: List[int],
        batch_size: Optional[int] = 50,
        save_source_date: Optional[bool] = True,
    ) -> None:
        """
        Copy annotations from one images IDs to another images IDs in API.

        :param src_image_ids: Images IDs in Supervisely.
        :type src_image_ids: List[int]
        :param dst_image_ids: Images IDs in Supervisely.
        :type dst_image_ids: List[int]
        :return: None
        :rtype: :class:`NoneType`
        :raises: :class:`RuntimeError` if len(src_image_ids) != len(dst_image_ids)

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_ids = [121236918, 121236919]
            dst_ids = [547837053, 547837054]
            api.annotation.copy_batch_by_ids(src_ids, dst_ids)
        """
        if len(src_image_ids) != len(dst_image_ids):
            raise RuntimeError(
                'Can not match "src_image_ids" and "dst_image_ids" lists, '
                "len(src_image_ids) != len(dst_image_ids)"
            )
        if len(src_image_ids) == 0:
            return
        for cur_batch in batched(list(zip(src_image_ids, dst_image_ids)), batch_size=batch_size):
            src_ids_batch, dst_ids_batch = zip(*cur_batch)
            self._api.post(
                "annotations.bulk.copy",
                data={
                    "srcImageIds": src_ids_batch,
                    "destImageIds": dst_ids_batch,
                    "preserveSourceDate": save_source_date,
                },
            )

    def _convert_json_info(self, info: dict, skip_missing=True) -> AnnotationInfo:
        """
        _convert_json_info
        """
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        return AnnotationInfo(**res._asdict())

    def append_labels(
        self,
        image_id: int,
        labels: List[Label],
        skip_bounds_validation: Optional[bool] = False,
    ) -> None:
        """
        Append labels to image with given ID in API.

        :param image_id: Image ID to append labels.
        :type image_id: int
        :param labels: List of labels to append.
        :type labels: List[Label]
        :return: None
        :rtype: :class:`NoneType`
        """
        if len(labels) == 0:
            return

        payload = []
        for label in labels:
            _label_json = label.to_json()
            _label_json["geometry"] = label.geometry.to_json()
            if "classId" not in _label_json:
                raise KeyError("Update project meta from server to get class id")
            payload.append(_label_json)

        added_ids = []
        for batch_jsons in batched(payload, batch_size=100):
            resp = self._api.post(
                "figures.bulk.add",
                {
                    ApiField.ENTITY_ID: image_id,
                    ApiField.FIGURES: batch_jsons,
                    ApiField.SKIP_BOUNDS_VALIDATION: skip_bounds_validation,
                },
            )
            for resp_obj in resp.json():
                figure_id = resp_obj[ApiField.ID]
                added_ids.append(figure_id)



# coding: utf-8

# docs
from requests import Response
from typing import List, NamedTuple, Dict, Optional, Callable, Union
from supervisely.task.progress import Progress
from tqdm import tqdm

from collections import defaultdict
from supervisely.api.module_api import ApiField, RemoveableBulkModuleApi
from supervisely.io.fs import ensure_base_path, get_file_hash
from supervisely._utils import batched

from supervisely.api.pointcloud.pointcloud_annotation_api import PointcloudAnnotationAPI
from supervisely.api.pointcloud.pointcloud_object_api import PointcloudObjectApi
from supervisely.api.pointcloud.pointcloud_figure_api import PointcloudFigureApi
from supervisely.api.pointcloud.pointcloud_tag_api import PointcloudTagApi
from requests_toolbelt import MultipartDecoder, MultipartEncoder


class PointcloudInfo(NamedTuple):
    """
    Object with :class:`Pointcloud<supervisely.pointcloud.pointcloud>` parameters from Supervisely.

    :Example:

    .. code-block:: python

        PointcloudInfo(
            id=19373403,
            frame=None,
            description='',
            name='000063.pcd',
            team_id=435,
            workspace_id=687,
            project_id=17231,
            dataset_id=55875,
            link=None,
            hash='7EcJCyhq15V4NnZ8oiPrKQckmXXypO4saqFN7kgH08Y=',
            path_original='/h5unms4-public/point_clouds/Z/h/bc/roHZP5nP2.pcd',
            cloud_mime='image/pcd',
            figures_count=4,
            objects_count=4,
            tags=[],
            meta={},
            created_at='2023-02-07T19:36:44.897Z',
            updated_at='2023-02-07T19:36:44.897Z'
        )
    """

    #: :class:`int`: Point cloud ID in Supervisely.
    id: int

    #: :class:`int`: Number of frame in the point cloud
    frame: int

    #: :class:`str`: Point cloud description.
    description: str

    #: :class:`str`: Point cloud filename.
    name: str

    #: :class:`int`: :class:`TeamApi<supervisely.api.team_api.TeamApi>` ID in Supervisely.
    team_id: int

    #: :class:`int`: :class:`WorkspaceApi<supervisely.api.workspace_api.WorkspaceApi>` ID in Supervisely.
    workspace_id: int

    #: :class:`int`: :class:`Project<supervisely.project.project.Project>` ID in Supervisely.
    project_id: int

    #: :class:`int`: :class:`Dataset<supervisely.project.project.Dataset>` ID in Supervisely.
    dataset_id: int

    #: :class:`str`: Link to point cloud.
    link: str

    #: :class:`str`: Point cloud hash obtained by base64(sha256(file_content)).
    #: Use hash for files that are expected to be stored at Supervisely or your deployed agent.
    hash: str

    #: :class:`str`: Relative storage URL to point cloud. e.g.
    #: "/h5un6l2bnaz1vms4-public/pointclouds/Z/d/HD/lfgipl...NXrg5vz.mp4".
    path_original: str

    #: :class:`str`: MIME type of the point cloud.
    cloud_mime: str

    #: :class:`int`: Number of PointcloudFigure objects in the point cloud
    figures_count: int

    #: :class:`int`: Number of PointcloudObject objects in the point cloud
    objects_count: int

    #: :class:`list`: Pointcloud :class:`PointcloudTag<supervisely.pointcloud_annotation.pointcloud_tag.PointcloudTag>` list.
    tags: list

    #: :class:`dict`: A dictionary containing point cloud metadata.
    meta: dict

    #: :class:`str`: Point cloud creation time. e.g. "2019-02-22T14:59:53.381Z".
    created_at: str

    #: :class:`str`: Time of last point cloud update. e.g. "2019-02-22T14:59:53.381Z".
    updated_at: str


class PointcloudApi(RemoveableBulkModuleApi):
    """
    API for working with :class:`Pointcloud<supervisely.pointcloud.pointcloud>`.
    :class:`PointcloudApi<PointcloudApi>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        pcd_id = 19618654
        pcd_info = api.pointcloud.get_info_by_id(pcd_id) # api usage example
    """

    def __init__(self, api):
        """
        :param api: Api class object
        """
        super().__init__(api)
        self.annotation = PointcloudAnnotationAPI(api)
        self.object = PointcloudObjectApi(api)
        self.figure = PointcloudFigureApi(api)
        self.tag = PointcloudTagApi(api)

    @staticmethod
    def info_sequence():
        """
        Get list of all :class:`PointcloudInfo<PointcloudInfo>` field names.

        :return: List of :class:`PointcloudInfo<PointcloudInfo>` field names.`
        :rtype: :class:`list`
        """

        return [
            ApiField.ID,
            ApiField.FRAME,
            ApiField.DESCRIPTION,
            ApiField.NAME,
            ApiField.TEAM_ID,
            ApiField.WORKSPACE_ID,
            ApiField.PROJECT_ID,
            ApiField.DATASET_ID,
            ApiField.LINK,
            ApiField.HASH,
            ApiField.PATH_ORIGINAL,
            # ApiField.PREVIEW,
            ApiField.CLOUD_MIME,
            ApiField.FIGURES_COUNT,
            ApiField.ANN_OBJECTS_COUNT,
            ApiField.TAGS,
            ApiField.META,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of :class:`PointcloudInfo<PointcloudInfo>` NamedTuple.

        :return: NamedTuple name.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            tuple_name = api.pointcloud.info_tuple_name()
            print(tuple_name) # PointCloudInfo
        """

        return "PointCloudInfo"

    def _convert_json_info(self, info: Dict, skip_missing: Optional[bool] = True):
        res = super(PointcloudApi, self)._convert_json_info(info, skip_missing=skip_missing)
        return PointcloudInfo(**res._asdict())

    def get_list(
        self,
        dataset_id: int,
        filters: Optional[List[Dict[str, str]]] = None,
    ) -> List[PointcloudInfo]:
        """
        Get list of information about all point cloud for a given dataset ID.

        :param dataset_id: :class:`Dataset<supervisely.project.project.Dataset>` ID in Supervisely.
        :type dataset_id: int
        :param filters: List of parameters to sort output Pointclouds. See: https://dev.supervise.ly/api-docs/#tag/Point-Clouds/paths/~1point-clouds.list/get
        :type filters: List[Dict[str, str]], optional
        :return: List of the point clouds objects from the dataset with given id.
        :rtype: :class:`List[PointcloudInfo]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 62664
            pcd_infos = api.pointcloud_episode.get_list(dataset_id)
            print(pcd_infos)
            # Output: [PointcloudInfo(...), PointcloudInfo(...)]

            id_list = [19618654, 19618657, 19618660]
            filtered_pointcloud_infos = api.pointcloud.get_list(dataset_id, filters=[{'field': 'id', 'operator': 'in', 'value': id_list}])
            print(filtered_pointcloud_infos)
            # Output:
            # [PointcloudInfo(id=19618654, ...), PointcloudInfo(id=19618657, ...), PointcloudInfo(id=19618660, ...)]
        """

        return self.get_list_all_pages(
            "point-clouds.list",
            {
                ApiField.DATASET_ID: dataset_id,
                ApiField.FILTER: filters or [],
            },
        )

    def get_info_by_id(self, id: int) -> PointcloudInfo:
        """
        Get point cloud information by ID in PointcloudInfo<PointcloudInfo> format.

        :param id: Point cloud ID in Supervisely.
        :type id: int
        :param raise_error: Return an error if the point cloud info was not received.
        :type raise_error: bool
        :return: Information about point cloud. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`PointcloudInfo`

        :Usage example:

         .. code-block:: python

            import supervisely as sly


            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pcd_id = 19373403
            pcd_info = api.pointcloud.get_info_by_id(pcd_id)
            print(pcd_info)

            # Output:
            # PointcloudInfo(
            #     id=19373403,
            #     frame=None,
            #     description='',
            #     name='000063.pcd',
            #     team_id=435,
            #     workspace_id=687,
            #     project_id=17231,
            #     dataset_id=55875,
            #     link=None,
            #     hash='7EcJCyhq15V4NnZ8oiPrKQckmXXypO4saqFN7kgH08Y=',
            #     path_original='/h5unms4-public/point_clouds/Z/h/bc/roHZP5nP2.pcd',
            #     cloud_mime='image/pcd',
            #     figures_count=4,
            #     objects_count=4,
            #     tags=[],
            #     meta={},
            #     created_at='2023-02-07T19:36:44.897Z',
            #     updated_at='2023-02-07T19:36:44.897Z'
            # )
        """
        return self._get_info_by_id(id, "point-clouds.info")

    def _download(self, id: int, is_stream: Optional[bool] = False):
        """
        :param id: int
        :param is_stream: bool
        :return: Response object containing pointcloud object with given id
        """
        response = self._api.post(
            "point-clouds.download",
            {ApiField.ID: id},
            stream=is_stream,
        )
        return response

    def download_path(self, id: int, path: str) -> None:
        """
        Download point cloud with given id on the given path.

        :param id: Point cloud ID in Supervisely.
        :type id: int
        :param path: Local save path for point cloud.
        :type path: str
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            storage_dir = sly.app.get_data_dir()
            pcd_id = 19373403
            pcd_info = api.pointcloud.get_info_by_id(pcd_id)
            save_path = os.path.join(storage_dir, pcd_info.name)

            api.pointcloud.download_path(pcd_id, save_path)
            print(os.listdir(storage_dir))

            # Output: ['000063.pcd']
        """

        response = self._download(id, is_stream=True)
        ensure_base_path(path)
        with open(path, "wb") as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)

    def get_list_related_images(self, id: int) -> List:
        """
        Get information about related context images.

        :param id: Point cloud ID in Supervisely.
        :type id: int
        :return: List of dictionaries with informations about related images
        :rtype: List
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pcd_id = 19373403
            img_infos = api.pointcloud.get_list_related_images(pcd_id)
            img_info = img_infos[0]
            print(img_info)

            # Output:
            # {
            #     'pathOriginal': '/h5un6qgms4-public/images/original/S/j/hJ/PwMg.png',
            #     'id': 473302,
            #     'entityId': 19373403,
            #     'createdAt': '2023-01-09T08:50:33.225Z',
            #     'updatedAt': '2023-01-09T08:50:33.225Z',
            #     'meta': {
            #         'deviceId': 'cam_2'},
            #         'fileMeta': {'mime': 'image/png',
            #         'size': 893783,
            #         'width': 1224,
            #         'height': 370
            #     },
            #     'hash': 'vxA+emfDNUkFP9P6oitMB5Q0rMlnskmV2jvcf47OjGU=',
            #     'link': None,
            #     'preview': '/previews/q/ext:jpeg/resize:fill:50:0:0/q:50/plain/h5ad-public/images/original/S/j/hJ/PwMg.png',
            #     'fullStorageUrl': 'https://dev.supervise.ly/hs4-public/images/original/S/j/hJ/PwMg.png',
            #     'name': 'img00'
            # }
        """

        dataset_id = self.get_info_by_id(id).dataset_id
        filters = [{"field": ApiField.ENTITY_ID, "operator": "=", "value": id}]
        return self.get_list_all_pages(
            "point-clouds.images.list",
            {ApiField.DATASET_ID: dataset_id, ApiField.FILTER: filters},
            convert_json_info_cb=lambda x: x,
        )

    def download_related_image(self, id: int, path: str) -> Response:
        """
        Download a related context image from Supervisely to local directory by image id.

        :param id: Related context imgage ID in Supervisely.
        :type id: int
        :param path: Local save path for point cloud.
        :type path: str
        :return: List of dictionaries with informations about related images
        :rtype: List
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            save_path = "src/output/img_0.png"
            img_info = api.pointcloud.get_list_related_images(pcd_info.id)[0]
            api.pointcloud.download_related_image(img_info["id"], save_path)
            print(f"Context image has been successfully downloaded to '{save_path}'")

        # Output:
        # Context image has been successfully downloaded to 'src/output/img_0.png'
        """

        response = self._api.post(
            "point-clouds.images.download",
            {ApiField.ID: id},
            stream=True,
        )
        ensure_base_path(path)
        with open(path, "wb") as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)
        return response

    # @TODO: copypaste from video_api
    def upload_hash(
        self,
        dataset_id: int,
        name: str,
        hash: str,
        meta: Optional[Dict] = None,
    ) -> PointcloudInfo:
        """
        Upload Pointcloud from given hash to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param name: Point cloud name.
        :type name: str
        :param hash: Point cloud hash.
        :type hash: str
        :param meta: Point cloud metadata.
        :type meta: dict, optional
        :return: Information about point cloud. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`PointcloudInfo`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dst_dataset_id = 62693

            src_pointcloud_id = 19618685
            pcd_info = api.pointcloud.get_info_by_id(src_pointcloud_id)
            hash, name, meta = pcd_info.hash, pcd_info.name, pcd_info.meta

            new_pcd_info = api.pointcloud.upload_hash(dst_dataset_id.id, name, hash, meta)
            print(new_pcd_info)

            # Output:
            # PointcloudInfo(
            #     id=19619507,
            #     frame=None,
            #     description='',
            #     name='0000000031.pcd',
            #     team_id=None,
            #     workspace_id=None,
            #     project_id=None,
            #     dataset_id=62694,
            #     link=None,
            #     hash='5w69Vv1i6JrqhU0Lw1UJAJFGPVWUzDG7O3f4QSwRfmE=',
            #     path_original='/j8a9qgms4-public/point_clouds/I/3/6U/L7YBY.pcd',
            #     cloud_mime='image/pcd',
            #     figures_count=None,
            #     objects_count=None,
            #     tags=None,
            #     meta={'frame': 31},
            #     created_at='2023-04-05T10:59:44.656Z',
            #     updated_at='2023-04-05T10:59:44.656Z'
            # )
        """

        meta = {} if meta is None else meta
        return self.upload_hashes(dataset_id, [name], [hash], [meta])[0]

    # @TODO: copypaste from video_api
    def upload_hashes(
        self,
        dataset_id: int,
        names: List[str],
        hashes: List[str],
        metas: Optional[List[Dict]] = None,
        progress_cb: Optional[Callable] = None,
    ) -> List[PointcloudInfo]:
        """
        Upload point clouds from given hashes to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param names: Point cloud name.
        :type names: List[str]
        :param hashes: Point cloud hash.
        :type hashes: List[str]
        :param metas: Point cloud metadata.
        :type metas: Optional[List[Dict]], optional
        :param progress_cb: Function for tracking upload progress.
        :type progress_cb: Progress, optional
        :return: List of informations about Pointclouds. See :class:`info_sequence<info_sequence>`
        :rtype: List[:class:`PointcloudInfo`]
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            src_dataset_id = 62664
            dst_dataset_id = 62690

            src_pcd_infos = api.pointcloud.get_list(src_dataset_id)

            names = [pcd.name for pcd in src_pcd_infos[:4]]
            hashes = [pcd.hash for pcd in src_pcd_infos[:4]]
            metas = [pcd.meta for pcd in src_pcd_infos[:4]]

            dst_pcd_infos = api.pointcloud.get_list(dst_dataset_id)
            print(f"{len(dst_pcd_infos)} pointcloud before upload.")
            # Output:
            # 0 pointcloud before upload.

            new_pcd_infos = api.pointcloud.upload_hashes(dst_dataset_id, names, hashes, metas)
            print(f"{len(new_pcd_infos)} pointcloud after upload.")
            # Output:
            # 4 pointcloud after upload.
        """

        return self._upload_bulk_add(
            lambda item: (ApiField.HASH, item), dataset_id, names, hashes, metas, progress_cb
        )

    # @TODO: copypaste from video_api
    def _upload_bulk_add(
        self,
        func_item_to_kv,
        dataset_id,
        names,
        items,
        metas=None,
        progress_cb=None,
    ):
        if metas is None:
            metas = [{}] * len(items)

        results = []
        if len(names) == 0:
            return results
        if len(names) != len(items):
            raise RuntimeError('Can not match "names" and "items" lists, len(names) != len(items)')

        for batch in batched(list(zip(names, items, metas))):
            images = []
            for name, item, meta in batch:
                item_tuple = func_item_to_kv(item)
                images.append(
                    {
                        ApiField.NAME: name,
                        item_tuple[0]: item_tuple[1],
                        ApiField.META: meta if meta is not None else {},
                    }
                )
            response = self._api.post(
                "point-clouds.bulk.add",
                {ApiField.DATASET_ID: dataset_id, ApiField.POINTCLOUDS: images},
            )
            if progress_cb is not None:
                progress_cb(len(images))

            results.extend([self._convert_json_info(item) for item in response.json()])
        name_to_res = {img_info.name: img_info for img_info in results}
        ordered_results = [name_to_res[name] for name in names]

        return ordered_results

    def upload_related_image(self, path: str) -> str:
        """
        Upload an image to the Supervisely. It generates us a hash for image.

        :param path: Image path.
        :type path: str
        :return: Hash for image. See :class:`info_sequence<info_sequence>`
        :rtype: str
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_file = src/input/img/000000.png"
            img_hash = api.pointcloud.upload_related_image(img_file)
            print(img_hash)

            # Output:
            # +R6dFy8nMEq6k82vHLxuakpqVBmyTTPj5hXdPfjAv/c=
        """

        return self.upload_related_images([path])[0]

    def upload_related_images(
        self,
        paths: List[str],
        progress_cb: Optional[Callable] = None,
    ) -> List[str]:
        """
        Upload a batch of related images to the Supervisely. It generates us a hashes for images.

        :param paths: Images pathes.
        :type paths: List[str]
        :return: List of hashes for images. See :class:`info_sequence<info_sequence>`
        :rtype: List[str]
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_paths = ["src/input/img/000001.png", "src/input/img/000002.png"]
            img_hashes = api.pointcloud.upload_related_images(img_paths)

            # Output:
            # [+R6dFy8nMEq6k82vHLxuakpqVBmyTTPjdfGdPfjAv/c=, +hfjbufnbkLhJb32vHLxuakpqVBmyTTPj5hXdPfhhj1c]
        """

        def path_to_bytes_stream(path):
            return open(path, "rb")

        return self._upload_data_bulk(path_to_bytes_stream, get_file_hash, paths, progress_cb)

    def add_related_images(
        self,
        images_json: List[Dict],
        camera_names: List[str] = None,
    ) -> Dict:
        """
        Attach images to point cloud.

        :param images_json: List of dictionaries with dataset id, image name, hash and meta.
        :type images_json: List[Dict]
        :param camera_names: List of camera informations.
        :type camera_names: List[Dict]
        :return: Response object
        :rtype: Dict
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_paths = ["src/input/img/000001.png", "src/input/img/000002.png"]
            cam_paths = ["src/input/cam_info/000001.json", "src/input/cam_info/000002.json"]

            img_hashes = api.pointcloud.upload_related_images(img_paths)
            img_infos = []
            for i, cam_info_file in enumerate(cam_paths):
                # reading cam_info
                with open(cam_info_file, "r") as f:
                    cam_info = json.load(f)
                img_info = {
                    "entityId": pcd_infos[i].id,
                    "name": f"img_{i}",
                    "hash": img_hashes[i],
                    "meta": cam_info,
                }
                img_infos.append(img_info)
            api.pointcloud.add_related_images(img_infos)
        """

        if camera_names is not None:
            if len(camera_names) != len(images_json):
                ValueError("camera_names length must be equal to images_json length.")
            for img_ind, camera_name in enumerate(camera_names):
                images_json[img_ind][ApiField.META]["deviceId"] = camera_name
        response = self._api.post("point-clouds.images.add", {ApiField.IMAGES: images_json})
        return response.json()

    def upload_path(
        self,
        dataset_id: int,
        name: str,
        path: str,
        meta: Optional[Dict] = None,
    ) -> PointcloudInfo:
        """
        Upload point cloud with given path to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param name: Point cloud name.
        :type name: str
        :param path: Path to point cloud.
        :type path: str
        :param meta: Dictionary with metadata for point cloud.
        :type meta: Optional[Dict]
        :return: Information about point cloud
        :rtype: PointcloudInfo
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pcd_file = "src/input/pcd/000000.pcd"
            pcd_info = api.pointcloud.upload_path(dataset.id, name="pcd_0", path=pcd_file)
            print(f'Point cloud "{pcd_info.name}" uploaded to Supervisely with ID:{pcd_info.id}')

            # Output:
            # Point cloud "pcd_0.pcd" uploaded to Supervisely with ID:19618685
        """

        metas = None if meta is None else [meta]
        return self.upload_paths(dataset_id, [name], [path], metas=metas)[0]

    def upload_paths(
        self,
        dataset_id: int,
        names: List[str],
        paths: List[str],
        progress_cb: Optional[Callable] = None,
        metas: Optional[List[Dict]] = None,
    ) -> List[PointcloudInfo]:
        """
        Upload point clouds with given paths to Dataset.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param names: Point clouds names.
        :type names: List[str]
        :param paths: Paths to point clouds.
        :type paths: List[str]
        :param progress_cb: Function for tracking upload progress.
        :type progress_cb: Progress, optional
        :param metas: List of dictionary with metadata for point cloud.
        :type metas: Optional[List[Dict]]
        :return: List of informations about point clouds
        :rtype: List[PointcloudInfo]
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            paths = ["src/input/pcd/000001.pcd", "src/input/pcd/000002.pcd"]
            pcd_infos = api.pointcloud.upload_paths(dataset.id, names=["pcd_1", "pcd_2"], paths=paths)
            print(f'Point clouds uploaded to Supervisely with IDs: {[pcd_info.id for pcd_info in pcd_infos]}')

            # Output:
            # Point clouds uploaded to Supervisely with IDs: [19618685, 19618686]
        """

        def path_to_bytes_stream(path):
            return open(path, "rb")

        hashes = self._upload_data_bulk(path_to_bytes_stream, get_file_hash, paths, progress_cb)
        return self.upload_hashes(dataset_id, names, hashes, metas=metas)

    def check_existing_hashes(self, hashes: List[str]) -> List[str]:
        """
        Check if point clouds with given hashes are exist.

        :param paths: Point clouds hashes.
        :type paths: List[str]
        :return: List of point clouds hashes.
        :rtype: List[str]
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pointcloud_id = 19618685
            pcd_info = api.pointcloud.get_info_by_id(pointcloud_id)
            hash = api.pointcloud.check_existing_hashes([pcd_info.hash])
            print(hash)

            # Output:
            # ['5w69Vv1i6JrqhU0Lw1UJAJFGPhgkIhs7O3f4QSwRfmE=']
        """

        results = []
        if len(hashes) == 0:
            return results
        for hashes_batch in batched(hashes, batch_size=900):
            response = self._api.post("images.internal.hashes.list", hashes_batch)
            results.extend(response.json())
        return results

    def _upload_data_bulk(self, func_item_to_byte_stream, func_item_hash, items, progress_cb):
        hashes = []
        if len(items) == 0:
            return hashes

        hash_to_items = defaultdict(list)

        for idx, item in enumerate(items):
            item_hash = func_item_hash(item)
            hashes.append(item_hash)
            hash_to_items[item_hash].append(item)

        unique_hashes = set(hashes)
        remote_hashes = self.check_existing_hashes(list(unique_hashes))
        new_hashes = unique_hashes - set(remote_hashes)

        if progress_cb is not None:
            progress_cb(len(remote_hashes))

        # upload only new images to supervisely server
        items_to_upload = []
        for hash in new_hashes:
            items_to_upload.extend(hash_to_items[hash])

        for batch in batched(items_to_upload):
            content_dict = {}
            for idx, item in enumerate(batch):
                content_dict["{}-file".format(idx)] = (
                    str(idx),
                    func_item_to_byte_stream(item),
                    "pcd/*",
                )
            encoder = MultipartEncoder(fields=content_dict)
            self._api.post("point-clouds.bulk.upload", encoder)
            if progress_cb is not None:
                progress_cb(len(batch))

        return hashes



# coding: utf-8
"""api for working with agent"""

from __future__ import annotations

from typing import NamedTuple, Optional, Dict, List
from enum import Enum
from supervisely.api.module_api import ApiField, ModuleApi, ModuleWithStatus


class AgentNotFound(Exception):
    """class AgentNotFound"""

    pass


class AgentNotRunning(Exception):
    """class AgentNotRunning"""

    pass


class AgentApi(ModuleApi, ModuleWithStatus):
    """
    API for working with agent. :class:`AgentApi<AgentApi>` object is immutable.

    :param api: API connection to the server
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        team_id = 8
        agents = api.agent.get_list(team_id)
    """

    class Status(Enum):
        """Agent status."""

        WAITING = "waiting"
        """"""
        RUNNING = "running"
        """"""

    @staticmethod
    def info_sequence():
        """
        NamedTuple AgentInfo information about Agent.

        :Example:

         .. code-block:: python

            AgentInfo("some info")
        """
        return [
            ApiField.ID,
            ApiField.NAME,
            ApiField.TOKEN,
            ApiField.STATUS,
            ApiField.USER_ID,
            ApiField.TEAM_ID,
            ApiField.CAPABILITIES,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        NamedTuple name - **AgentInfo**.
        """
        return "AgentInfo"

    def __init__(self, api):
        ModuleApi.__init__(self, api)
        ModuleWithStatus.__init__(self)

    def get_list(
        self, team_id: int, filters: Optional[List[Dict[str, str]]] = None
    ) -> List[NamedTuple]:
        """
        List of all agents in the given Team.

        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param filters: List of params to sort output Agents.
        :type filters: List[dict], optional
        :return: List of Agents with information. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            team_id = 16087
            agents = api.agent.get_list(team_id)

            filter_agents = api.agent.get_list(team_id, filters=[{ 'field': 'name', 'operator': '=', 'value': 'Gorgeous Chicken' }])
        """
        return self.get_list_all_pages("agents.list", {"teamId": team_id, "filter": filters or []})

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Agent information by ID.

        :param id: Agent ID in Supervisely.
        :type id: int
        :return: Information about Agent. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            agent = api.agent.get_info_by_id(7)
        """
        return self._get_info_by_id(id, "agents.info")

    def get_status(self, id: int) -> AgentApi.Status:
        """
        Status object containing status of Agent: waiting or running.

        :param id: Agent ID in Supervisely.
        :type id: int
        :return: Agent Status
        :rtype: :class:`Status<supervisely.api.agent_api.AgentApi.Status>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            agent = api.agent.get_status(7)
        """
        status_str = self.get_info_by_id(id).status
        return self.Status(status_str)

    def raise_for_status(self, status):
        """raise_for_status"""
        pass



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from typing import Optional


class Icons(Widget):
    def __init__(
        self,
        class_name: Optional[str] = None,
        color: Optional[str] = None,
        bg_color: Optional[str] = None,
        rounded: Optional[bool] = False,
        image_url: Optional[str] = None,
        widget_id: Optional[str] = None,
    ):
        self._class_name = class_name
        self._color = color
        self._bg_color = bg_color
        self._rounded = rounded
        self._image_url = image_url

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "className": self._class_name,
            "color": self._color,
            "bgColor": self._bg_color,
            "rounded": self._rounded,
            "imageUrl": self._image_url,
        }

    def get_json_state(self):
        return {}

    def set_class_name(self, value: str):
        self._class_name = value
        DataJson()[self.widget_id]["className"] = self._class_name
        DataJson().send_changes()

    def get_class_name(self):
        self._class_name = DataJson()[self.widget_id]["className"]
        return self._class_name

    def set_color(self, value: str):
        self._color = value
        DataJson()[self.widget_id]["color"] = self._color
        DataJson().send_changes()

    def get_color(self):
        self._color = DataJson()[self.widget_id]["color"]
        return self._color

    def set_bg_color(self, value: str):
        self._bg_color = value
        DataJson()[self.widget_id]["bgColor"] = self._bg_color
        DataJson().send_changes()

    def get_bg_color(self):
        self._bg_color = DataJson()[self.widget_id]["bgColor"]
        return self._bg_color

    def set_rounded(self):
        self._rounded = True
        DataJson()[self.widget_id]["rounded"] = self._rounded
        DataJson().send_changes()

    def set_standard(self):
        self._rounded = False
        DataJson()[self.widget_id]["rounded"] = self._rounded
        DataJson().send_changes()

    def set_image_url(self, value: str):
        self._image_url = value
        DataJson()[self.widget_id]["imageUrl"] = self._image_url
        DataJson().send_changes()

    def get_image_url(self):
        self._image_url = DataJson()[self.widget_id]["imageUrl"]
        return self._image_url



# coding: utf-8
from typing import Optional, Callable, Union

from tqdm import tqdm

from supervisely.api.module_api import ModuleApiBase, ApiField
from supervisely.collection.str_enum import StrEnum
from supervisely.io.fs import ensure_base_path, get_file_name_with_ext
from requests_toolbelt import MultipartEncoder
import mimetypes


class Provider(StrEnum):
    """Provider"""

    S3 = "s3"
    """S3"""
    GOOGLE = "google"
    """GOOGLE"""
    AZURE = "azure"
    """AZURE"""
    FS = "fs"
    """FS"""

    @staticmethod
    def validate_path(path):
        if (
            not path.startswith(str(Provider.S3))
            and not path.startswith(str(Provider.GOOGLE))
            and not path.startswith(str(Provider.AZURE))
            and not path.startswith(str(Provider.FS))
        ):
            raise ValueError(
                "Incorrect cloud path, learn more here: https://docs.supervise.ly/enterprise-edition/advanced-tuning/s3#links-plugin-cloud-providers-support"
            )


class RemoteStorageApi(ModuleApiBase):
    """RemoteStorageApi"""

    def _convert_json_info(self, info: dict):
        """_convert_json_info"""
        return info

    def get_file_info_by_path(
        self,
        path: str,
    ) -> dict:
        """
        Get info about file for given remote path.

        :param path: Remote path to file.
        :type path: str
        :returns: List of files in the given remote path
        :rtype: dict

        """

        Provider.validate_path(path)
        path = path.rstrip("/")
        resp = self._api.get(
            "remote-storage.list",
            {
                ApiField.PATH: path,
                "recursive": False,
                "files": True,
                "folders": False,
                "limit": 1,
                "startAfter": "",
            },
        )
        return resp.json()[0]

    def list(
        self,
        path: str,
        recursive: bool = True,
        files: bool = True,
        folders: bool = True,
        limit: int = 10000,
        start_after: str = "",
    ) -> list:
        """
        List files and directories for given remote path.

        :param path: Remote path with items that you want to list.
        :type path: str
        :param recursive: List remote path revursively.
        :type recursive: bool
        :param files: List files in the given path.
        :type files: bool
        :param folders: List folders in the given path.
        :type folders: bool
        :param limit: Limit of files to list. 10000 is the maximum limit.
        :type limit: int
        :param start_after: Start listing path after given file name.
        :type start_after: str
        :returns: List of files in the given remote path
        :rtype: dict

        """

        Provider.validate_path(path)
        path = path.rstrip("/") + "/"
        resp = self._api.get(
            "remote-storage.list",
            {
                ApiField.PATH: path,
                "recursive": recursive,
                "files": files,
                "folders": folders,
                "limit": limit,
                "startAfter": start_after,
            },
        )
        return resp.json()

    def download_path(
        self, remote_path: str, save_path: str, progress_cb: Optional[Union[tqdm, Callable]] = None
    ):
        """
        Downloads item from given remote path to given local path.

        :param remote_path: Remote path to item that you want to download.
        :type remote_path: str
        :param save_path: Local save path.
        :type save_path: str
        :param progress_cb: Progress function to download.
        :type progress_cb: tqdm or callable, optional


        .. code-block:: python

            provider = "s3" # can be one of ["s3", "google", "azure"]
            bucket = "bucket-test-export"
            path_in_bucket = "/demo/image.jpg"
            remote_path = api.remote_storage.get_remote_path(provider, bucket, path_in_bucket)
            # or alternatively use this:
            # remote_path = f"{provider}://{bucket}{path_in_bucket}"
            api.remote_storage.upload_path(local_path="images/my-cats.jpg", remote_path=remote_path)
        """
        Provider.validate_path(remote_path)
        ensure_base_path(save_path)
        response = self._api.post(
            "remote-storage.download", {ApiField.LINK: remote_path}, stream=True
        )
        # if "Content-Length" in response.headers:
        #     length = int(response.headers['Content-Length'])
        with open(save_path, "wb") as fd:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                fd.write(chunk)
                if progress_cb is not None:
                    progress_cb(len(chunk))

    def upload_path(self, local_path: str, remote_path: str):
        """
        Uploads item from given local path to given remote path.

        :param local_path: Local path to item that you want to upload.
        :type local_path: str
        :param remote_path: Remote destination path.
        :type remote_path: str
        :Usage example:

        .. code-block:: python

            provider = "s3" # can be one of ["s3", "google", "azure"]
            bucket = "bucket-test-export"
            path_in_bucket = "/demo/image.jpg"
            remote_path = api.remote_storage.get_remote_path(provider, bucket, path_in_bucket)
            # or alternatively use this:
            # remote_path = f"{provider}://{bucket}{path_in_bucket}"
            api.remote_storage.upload_path(local_path="images/my-cats.jpg", remote_path=remote_path)
        """
        Provider.validate_path(remote_path)
        return self._upload_paths_batch([local_path], [remote_path])

    def _upload_paths_batch(self, local_paths, remote_paths):
        """_upload_paths_batch"""

        if len(local_paths) != len(remote_paths):
            raise ValueError("Inconsistency in paths, len(local_paths) != len(remote_paths)")
        if len(local_paths) == 0:
            return {}

        def path_to_bytes_stream(path):
            return open(path, "rb")

        content = []
        for idx, (src, dst) in enumerate(zip(local_paths, remote_paths)):
            content.append((ApiField.PATH, dst))
            name = get_file_name_with_ext(dst)
            content.append(
                (
                    "file",
                    (
                        name,
                        path_to_bytes_stream(src),
                        mimetypes.MimeTypes().guess_type(src)[0],
                    ),
                )
            )
        encoder = MultipartEncoder(fields=content)
        resp = self._api.post("remote-storage.upload", encoder)
        return resp.json()

    def get_remote_path(self, provider: str, bucket: str, path_in_bucket: str) -> str:
        """
        Returns remote path.

        :param provider: Can be one of "s3", "google", "azure".
        :type provider: str
        :param bucket: Name of the bucket container.
        :type bucket: str
        :param path_in_bucket: Path to item in bucket.
        :type path_in_bucket: str
        :Usage example:

        .. code-block:: python

            provider = "s3"
            bucket = "bucket-test-export"
            path_in_bucket = "/demo/image.jpg"
            remote_path = api.remote_storage.get_remote_path(provider, bucket, path_in_bucket)
            # Output: s3://bucket-test-export/demo/image.jpg
        """
        res_path = f"{provider}://{bucket}/{path_in_bucket.lstrip('/')}"
        Provider.validate_path(res_path)
        return res_path
